
import pytest

from coinmetrics.api_client import CoinMetricsClient
import os

client = CoinMetricsClient(str(os.environ.get("CM_API_KEY")), verbose=True)
cm_api_key_set = os.environ.get("CM_API_KEY") is not None
REASON_TO_SKIP = "Need to set CM_API_KEY as an env var in order to run this test"

print("CM_API_KEY is set - tests will run") if cm_api_key_set else print(
    "CM_API_KEY not set, tests will not run"
)

@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_constituent_snapshots() -> None:
    data = client.get_snapshots_of_asset_metric_constituents(metric="volume_trusted_spot_usd_1d").to_list()
    assert all(["time" in item for item in data])
    assert all(["exchange" in item for item in data])


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_constituent_timeframes() -> None:
    data = client.get_timeframes_of_asset_metric_constituents(metric="volume_trusted_spot_usd_1d").to_list()
    assert all(["exchange" in item for item in data])
    assert all(["start_time" in item for item in data])
    assert all(["end_time" in item for item in data])






import os
import threading
import pytest

from coinmetrics.api_client import CoinMetricsClient

client = CoinMetricsClient(str(os.environ.get("CM_API_KEY")))
cm_api_key_set = os.environ.get("CM_API_KEY") is not None
REASON_TO_SKIP = "Need to set CM_API_KEY as an env var in order to run this test"

print("CM_API_KEY is set - tests will run") if cm_api_key_set else print(
    "CM_API_KEY not set, tests will not run"
)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_security_master_markets() -> None:
    expected_cols = ["market", "exchange", "type"]
    markets = client.security_master_markets().first_page()[0]
    assert all([col in markets for col in expected_cols])


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_security_master_assets() -> None:
    expected_cols = ["asset"]
    assets = client.security_master_assets().first_page()[0]
    assert all([col in assets for col in expected_cols])


if __name__ == '__main__':
    pytest.main()


import os
import pytest

from coinmetrics.api_client import CoinMetricsClient

client = CoinMetricsClient(str(os.environ.get("CM_API_KEY")))
cm_api_key_set = os.environ.get("CM_API_KEY") is not None
REASON_TO_SKIP = "Need to set CM_API_KEY as an env var in order to run this test"

print("CM_API_KEY is set - tests will run") if cm_api_key_set else print(
    "CM_API_KEY not set, tests will not run"
)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_security_master_markets() -> None:
    required_market_cols = [
        "market",
        "exchange",
        "type"
    ]
    security_master_market = client.security_master_markets().first_page()[0]
    assert all([col in security_master_market for col in required_market_cols])


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_security_master_assets() -> None:
    security_master_asset = client.security_master_assets(assets="btc").first_page()[0]
    assert 'asset' in security_master_asset
    assert security_master_asset.get("type") == "utxo"



import datetime

import pytest

from coinmetrics.api_client import CoinMetricsClient
import os

client = CoinMetricsClient(str(os.environ.get("CM_API_KEY")), verbose=True)
cm_api_key_set = os.environ.get("CM_API_KEY") is not None
REASON_TO_SKIP = "Need to set CM_API_KEY as an env var in order to run this test"

print("CM_API_KEY is set - tests will run") if cm_api_key_set else print(
    "CM_API_KEY not set, tests will not run"
)

@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_get_full_transaction_for_block() -> None:
    asset = "btc"
    block_hash = "0000000000000000000334e8637314d72d86a533c71f48da23e85e70a82cd38a"
    transaction_id = "29d401526b06d55749034c10c3ee7ffd9ecab942c9b6852c963fa61103552729"
    transaction = client.get_full_transaction_for_block(
        asset=asset, block_hash=block_hash, txid=transaction_id
    )
    as_list = transaction
    print(as_list)
    assert len(as_list) >= 1


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_get_blockchain_list_of_account_balance_updates() -> None:
    asset = "btc"
    account = "112jmDkNGHSbhhY17JGpxU3sMA9ZExG7b2"
    updates = client.get_list_of_balance_updates_for_account_v2(asset=asset, account=account).to_list()
    assert updates[0]['block_hash'] == "0000000000000000002c7505ef2272e0677fa53d68d633f8e076ed42dd3380e6"


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_get_full_block_v1() -> None:
    asset = "eth"
    block_hash = "0x27a2bd0fd3b3298dd8004c18aaad83374bdc5dbd36eac46bfe00772d88dba7cf"
    full_block = client.get_full_block(asset=asset, block_hash=block_hash)
    assert full_block['block_hash'] == block_hash.strip("0x")
    assert full_block['height'] == '15442095'


if __name__ == "__main__":
    pytest.main()



from typing import Optional, List, Dict
from dataclasses import dataclass, field


@dataclass
class CoinMetricsAPIModel:

    @classmethod
    def get_dataframe_cols(cls) -> List[str]:
        return list(cls.__annotations__.keys())


@dataclass
class AssetChainsData(CoinMetricsAPIModel):
    """
    Class to represent data returned from asset chains data type
    """
    asset: str
    time: str
    chains_count: str
    blocks_count_at_tip: str
    reorg: Optional[str] = field(default="false")
    reorg_depth: str = field(default="0")
    chains: List[List[Dict[str, str]]] = field(default_factory=list)


@dataclass
class TransactionTrackerData(CoinMetricsAPIModel):
    """
    Class to represent data returned from tranasaction-tracker endpoint
    """
    txid: str
    time: str
    first_seen_time: str
    status: str
    status_update_time: str
    status_updates: List[str]
    details: str
    height: str = field(default="0")
    mempool_approx_queue_position: str = field(default="None")
    next_block_approx_settlement_probability_pct: str = field(default="0")
    block_hash: str = field(default="None")
    geo: str = field(default="None")
    replacement_for_txid: str = field(default="None")
    inputs: List[str] = field(default_factory=list)
    outputs: List[str] = field(default_factory=list)



from __future__ import annotations

import os
import warnings

import requests
import itertools

from dateutil.relativedelta import relativedelta
from copy import deepcopy
from gzip import GzipFile
from io import BytesIO
from logging import getLogger
from time import sleep
from datetime import datetime, timedelta, date
from typing import Any, Dict, Iterable, Iterator, List, Optional, cast, Type, Callable, Union, Generator, Tuple
from dateutil.parser import isoparse
from coinmetrics._typing import (
    DataRetrievalFuncType,
    DataReturnType,
    FilePathOrBuffer,
    UrlParamTypes,
    DataFrameType,
)
from coinmetrics._utils import get_file_path_or_buffer
from coinmetrics._models import AssetChainsData, CoinMetricsAPIModel, TransactionTrackerData
from importlib import import_module
from concurrent.futures import ThreadPoolExecutor, Executor
from tqdm import tqdm
from collections import defaultdict

orjson_found = True
try:
    json = import_module("orjson")
except ModuleNotFoundError:
    orjson_found = False

if not orjson_found:
    import json
else:
    import orjson as json

isoparse_typed: Callable[[Union[str, bytes]], datetime] = isoparse


logger = getLogger("cm_client_data_collection")

try:
    import pandas as pd
except ImportError:
    logger.info(
        "Pandas export is unavailable. Install pandas to unlock dataframe functions."
    )


class CsvExportError(Exception):
    pass


class DataFetchError(Exception):
    pass


NUMBER_OF_RETRIES = 3


class DataCollection:

    API_RETURN_MODEL: Optional[Type[CoinMetricsAPIModel]] = None

    def __init__(
        self,
        data_retrieval_function: DataRetrievalFuncType,
        endpoint: str,
        url_params: Dict[str, UrlParamTypes],
        csv_export_supported: bool = True,
    ) -> None:
        self._csv_export_supported = csv_export_supported
        self._data_retrieval_function = data_retrieval_function
        self._endpoint = endpoint
        self._url_params = url_params
        self._next_page_token: Optional[str] = ""
        self._current_data_iterator: Optional[Iterator[Any]] = None

    def first_page(self) -> List[Dict[str, Any]]:
        return cast(
            List[Dict[str, Any]],
            self._fetch_data_with_retries(self._url_params)["data"],
        )

    def to_list(self) -> List[Dict[str, Any]]:
        return [data for data in self]

    def __next__(self) -> Any:
        try:
            if self._current_data_iterator is not None:
                return next(self._current_data_iterator)
        except StopIteration:
            if self._next_page_token is None:
                raise StopIteration

        url_params = deepcopy(self._url_params)
        if self._next_page_token:
            url_params["next_page_token"] = self._next_page_token
        api_response = self._data_retrieval_function(self._endpoint, url_params)
        self._next_page_token = cast(Optional[str], api_response.get("next_page_token"))
        self._current_data_iterator = iter(api_response["data"])
        return next(self._current_data_iterator)

    def __iter__(self) -> "DataCollection":
        return self

    def _fetch_data_with_retries(
        self, url_params: Dict[str, UrlParamTypes]
    ) -> DataReturnType:
        for i in range(1, NUMBER_OF_RETRIES + 1):
            try:
                return self._data_retrieval_function(self._endpoint, url_params)
            except requests.exceptions.ProxyError as e:
                if i == NUMBER_OF_RETRIES:
                    raise
                logger.warning(
                    "failed to fetch data with error: %s, retrying in 1 second, try (%s/%s)",
                    e,
                    i,
                    NUMBER_OF_RETRIES,
                )
                sleep(1)

        raise DataFetchError("Failed to fetch data")

    def export_to_csv(
        self,
        path_or_bufstr: FilePathOrBuffer = None,
        columns_to_store: Optional[List[str]] = None,
        compress: bool = False,
    ) -> Optional[str]:
        if not self._csv_export_supported:
            raise CsvExportError(
                "Sorry, csv export is not supported for this data type."
            )

        return self._export_to_file(
            self._get_csv_data_lines(columns_to_store), path_or_bufstr, compress
        )

    def _get_csv_data_lines(
        self, columns_to_store: Optional[List[str]]
    ) -> Iterable[bytes]:
        first_data_el = None
        if columns_to_store is None:
            try:
                first_data_el = next(self)
            except StopIteration:
                logger.info("no data to export")
                return
            if self.API_RETURN_MODEL:
                columns_to_store = self.API_RETURN_MODEL.get_dataframe_cols()
            else:
                columns_to_store = list(first_data_el.keys())

        yield (",".join(columns_to_store) + "\n").encode()

        if first_data_el is not None:
            yield (
                ",".join(
                    f'"{first_data_el.get(column)}"' or ""
                    for column in columns_to_store
                )
                + "\n"
            ).encode()
        for data_el in self:
            yield (
                ",".join(
                    f'"{data_el.get(column)}"' or "" for column in columns_to_store
                )
                + "\n"
            ).encode()

    def export_to_json(
        self,
        path_or_bufstr: FilePathOrBuffer = None,
        compress: bool = False,
    ) -> Optional[str]:
        def _gen_json_lines() -> Iterable[bytes]:
            for data_row in self:
                yield json.dumps(data_row) + b"\n"

        return self._export_to_file(_gen_json_lines(), path_or_bufstr, compress)

    def _export_to_file(
        self,
        data_generator: Iterable[bytes],
        path_or_bufstr: FilePathOrBuffer = None,
        compress: bool = False,
    ) -> Optional[str]:

        if path_or_bufstr is None:
            path_or_bufstr_obj: FilePathOrBuffer = BytesIO()
        else:
            path_or_bufstr_obj = path_or_bufstr

        path_or_bufstr_obj = get_file_path_or_buffer(path_or_bufstr_obj)
        if hasattr(path_or_bufstr_obj, "write"):
            f = path_or_bufstr_obj
            close = False
        else:
            f = open(path_or_bufstr_obj, "wb")  # type: ignore
            close = True
        if compress:
            output_file = GzipFile(fileobj=f)  # type: ignore
        else:
            output_file = f  # type: ignore
        try:
            for line in data_generator:
                output_file.write(line)
        finally:
            if compress:
                output_file.close()
            if close:
                f.close()  # type: ignore

        if path_or_bufstr is None:
            return path_or_bufstr_obj.getvalue().decode()  # type: ignore

        return None

    def to_dataframe(
        self,
        header: Optional[List[str]] = None,
        dtype_mapper: Optional[Dict[str, Any]] = None,
        optimize_pandas_types: Optional[bool] = True,
    ) -> DataFrameType:
        """
        Outputs a pandas dataframe.

        :param header: Optional column names for outputted dataframe. List length must match the output.
        :type header: list(str)
        :param dtype_mapper: Dictionary for converting columns to types, where keys are columns and values are types that pandas accepts in an `as_type()` call. This mapping is prioritized over the pandas dtype conversions.
        :type dtype_mapper: dict
        :param optimize_pandas_types: Boolean flag for using pandas typing for dataframes. If True and dtype_mapper is not specified, then the output will result in all columns with string dtype. If True and dtype_mapper is specified, then the output will convert the dataframe columns with the user-specified dtyoes.
        :type optimize_pandas_types: Bool
        :return: Data in a pandas dataframe
        :rtype: DataFrameType
        """
        if pd is None:
            logger.info("Pandas not found; Returning None")
            return None
        else:
            if optimize_pandas_types:
                f = BytesIO()
                self.export_to_csv(f)
                if f.getbuffer().nbytes == 0:
                    return pd.DataFrame()
                else:
                    f.seek(0)
                    # if self.API_RETURN_MODEL:
                    #     columns = self.API_RETURN_MODEL.get_dataframe_cols()
                    # else:
                    columns = (
                        BytesIO(f.getvalue())
                        .readlines(1)[0]
                        .decode()
                        .strip()
                        .split(",")
                    )
                    datetime_cols = [
                        c for c in columns if c.endswith("_time") or c == "time"
                    ]
                    buffer: BytesIO = f
                    cols: List[str] = datetime_cols
                    dtype_map: Optional[Dict[str, Any]] = dtype_mapper
                    df: pd.DataFrame = pd.read_csv(
                        buffer,
                        parse_dates=cols,
                        dtype=dtype_map,
                    )
                    if dtype_mapper is None:
                        df = df.convert_dtypes()
                    if header is not None:
                        assert len(df.columns) == len(
                            header
                        ), "header length does not match output values"
                        df.columns = pd.Index(header)
                    return df
            else:
                if dtype_mapper is None:
                    return pd.DataFrame(self)
                else:
                    return pd.DataFrame(self).astype(dtype=dtype_mapper)

    def parallel(self,
                 parallelize_on: Optional[Union[str, List[str]]] = None,
                 executor: Optional[Callable[[Any], Executor]] = None,
                 max_workers: Optional[int] = None,
                 progress_bar: Optional[bool] = None,
                 time_increment: Optional[Union[relativedelta, timedelta]] = None
                 ) -> "ParallelDataCollection":
        """
        This method will convert the DataCollection into a ParallelDataCollection - enabling the ability to split
        one http request into many HTTP requests for faster data export. By default this will be split based on the
        primary query parameter. For example if you query get_asset_metrics(assets=....) it will split into many requests
        based on the assets.
        :param parallelize_on: parameter(s) to parallelize on. Can be any list type parameters
        :type parallelize_on: List[str], str
        :param executor: By defualt the ParallelDataCollection will use a ProcessPoolExecutor, but this can be changed
        :type executor: Executor
        :param max_workers: Specify the number of parallel threads. By default this is 10 and cannot be increased beyond 10
        :type: int
        :param progress_bar: flag to show a progress bar for data export or not, by default is true
        :type progress_bar: bool
        :param time_increment: option to parallelize by a time. Can use timedelta for time periods in weeks and relativedelta for longer time periods like a month or year
        :type time_increment: timedelta, relativedelta
        :return: ParallelDataCollection that matches the existing one
        """
        return ParallelDataCollection(self,
                                      parallelize_on=parallelize_on,
                                      executor=executor,
                                      max_workers=max_workers,
                                      progress_bar=progress_bar,
                                      time_increment=time_increment)


class AssetChainsDataCollection(DataCollection):

    API_RETURN_MODEL = AssetChainsData

    def to_dataframe(
        self,
        header: Optional[List[str]] = None,
        dtype_mapper: Optional[Dict[str, Any]] = None,
        optimize_pandas_types: Optional[bool] = True,
    ) -> DataFrameType:
        df = super().to_dataframe(header=header, dtype_mapper=dtype_mapper, optimize_pandas_types=optimize_pandas_types)
        if 'reorg' in df.columns:
            df['reorg'] = df['reorg'].apply(lambda reorg: True if reorg == "true" else False)
        return df


class TransactionTrackerDataCollection(DataCollection):

    API_RETURN_MODEL = TransactionTrackerData


class CatalogV2DataCollection(DataCollection):
    """
    This class will be used to implement functionality specific to catalog-v2 endpoints.
    """


class ParallelDataCollection(DataCollection):
    """
    This class will be used as an extension of the normal data collection, but all functions will run in parallel,
    utilizing Python's concurrent.futures features. The main purpose of this class is for historical export of
    data.
    """

    TIME = "time"
    _VALID_PARALLELIZATION_PARAMS = {'exchanges', 'assets', 'indexes', 'metrics', 'markets', 'institutions',
                                     'defi_protocols', 'exchange_assets', 'pairs'}
    _ENDPOINT_FIRST_PARAM_DICT = {
        'blockchain-metadata/tags': 'type',
        'blockchain-v2/{asset}/accounts': 'asset',
        'blockchain-v2/{asset}/accounts/{account}/balance-updates': 'asset',
        'blockchain-v2/{asset}/balance-updates': 'asset',
        'blockchain-v2/{asset}/blocks': 'asset',
        'blockchain-v2/{asset}/blocks/{block_hash}': 'asset',
        'blockchain-v2/{asset}/blocks/{block_hash}/transactions/{txid}': 'asset',
        'blockchain-v2/{asset}/sub-accounts': 'asset',
        'blockchain-v2/{asset}/transactions': 'asset',
        'blockchain-v2/{asset}/transactions/{txid}': 'asset',
        'blockchain/{asset}/accounts': 'asset',
        'blockchain/{asset}/balance-updates': 'asset',
        'blockchain/{asset}/blocks': 'asset',
        'blockchain/{asset}/blocks/{block_hash}': 'asset',
        'blockchain/{asset}/blocks/{block_hash}/transactions/{txid}': 'asset',
        'blockchain/{asset}/settlement': 'asset',
        'blockchain/{asset}/transaction-tracker': 'asset',
        'blockchain/{asset}/transactions': 'asset',
        'blockchain/{asset}/transactions/{txid}': 'asset',
        'catalog-v2/asset-metrics': 'assets',
        'catalog-v2/exchange-asset-metrics': 'exchange_assets',
        'catalog-v2/exchange-metrics': 'exchanges',
        'catalog-v2/institution-metrics': 'institutions',
        'catalog-v2/market-candles': 'markets',
        'catalog-v2/market-contract-prices': 'markets',
        'catalog-v2/market-funding-rates': 'markets',
        'catalog-v2/market-greeks': 'markets',
        'catalog-v2/market-implied-volatility': 'markets',
        'catalog-v2/market-liquidations': 'markets',
        'catalog-v2/market-metrics': 'markets',
        'catalog-v2/market-openinterest': 'markets',
        'catalog-v2/market-orderbooks': 'markets',
        'catalog-v2/market-quotes': 'markets',
        'catalog-v2/market-trades': 'markets',
        'catalog-v2/index-candles': 'indexes',
        'catalog-v2/index-levels': 'indexes',
        'catalog-v2/pair-metrics': 'pairs',
        'profile/assets': 'assets',
        'taxonomy-metadata/assets': 'version',
        'taxonomy/assets': 'assets',
        'timeseries/asset-alerts': 'assets',
        'timeseries/asset-chains': 'assets',
        'timeseries/asset-metrics': 'assets',
        'timeseries/defi-balance-sheets': 'defi_protocols',
        'timeseries/exchange-asset-metrics': 'exchange_assets',
        'timeseries/exchange-metrics': 'exchanges',
        'timeseries/index-candles': 'indexes',
        'timeseries/index-constituents': 'indexes',
        'timeseries/index-levels': 'indexes',
        'timeseries/institution-metrics': 'institutions',
        'timeseries/market-candles': 'markets',
        'timeseries/market-contract-prices': 'markets',
        'timeseries/market-funding-rates': 'markets',
        'timeseries/market-greeks': 'markets',
        'timeseries/market-implied-volatility': 'markets',
        'timeseries/market-liquidations': 'markets',
        'timeseries/market-metrics': 'markets',
        'timeseries/market-openinterest': 'markets',
        'timeseries/market-orderbooks': 'markets',
        'timeseries/market-quotes': 'markets',
        'timeseries/market-trades': 'markets',
        'timeseries/mempool-feerates': 'assets',
        'timeseries/mining-pool-tips-summary': 'assets',
        'timeseries/pair-candles': 'pairs',
        'timeseries/pair-metrics': 'pairs',
        'reference-data/markets': "markets",
        'reference-data/asset-metrics': 'metrics',
        'reference-data/exchange-metrics': 'metrics',
        'reference-data/exchange-asset-metrics': 'metrics',
        'reference-data/pair-metrics': 'metrics',
        'reference-data/institution-metrics': 'metrics'}

    def __init__(self, parent_data_collection: DataCollection,
                 parallelize_on: Optional[Union[str, List[str]]] = None,
                 executor: Optional[Callable[..., Executor]] = None,
                 max_workers: Optional[int] = None,
                 progress_bar: Optional[bool] = None,
                 time_increment: Optional[Union[relativedelta, timedelta]] = None):
        """
        :param parallelize_on: What parameter to parallelize on. By default will use the primary query parameter in the
        endpoint the user is calling. For example - if the user is calling `.get_market_candles(assets="...") it will
        split their request into many separate requests, one for each asset
        :param executor: by default this class uses ProcessPoolExecutor for concurrency, this could be swapped out for
        ThreadPoolExecutor or something else custom, based on User needs
        :param max_workers: The default max_workers number is 10 - so up to 10 processes or threads will be running at
        once. Increasing this can make the code run faster, but users may run into issues with resources, or start to hit
        rate limits.
        :param progress_bar: By default this class uses a tqdm progress bar to show the progress of the threads finishing
        so it is clear what is happening during long running intervals. Can be set to false to disable
        :param time_increment: Optionally, can split the data collections by time_increment. This feature splits
        data collections further by time increment. So if you split by MONTH this will split a year long request into
        12 smaller requests. If there is no "start_time" in the request it will raise a ValueError
        """
        super().__init__(parent_data_collection._data_retrieval_function, parent_data_collection._endpoint,
                         parent_data_collection._url_params, parent_data_collection._csv_export_supported)
        self._parallelize_on = self._get_parallelize_on(parallelize_on)
        self._executor: Callable[..., Executor] = executor if executor else ThreadPoolExecutor  # type: ignore
        self._max_workers = max_workers if max_workers else 10
        if self._max_workers > 10:
            warnings.warn("Max workers greater than 10 are not permitted due to rate limits restrictions")
            self._max_workers = 10
        self._progress_bar = progress_bar if progress_bar is not None else True
        self._time_increment = time_increment
        if self._time_increment is not None:
            self._url_params.update({"end_inclusive": False})

    def get_parallel_datacollections(self) -> List[DataCollection]:
        """
        This method creates a list of data collections all possible combinations of all the url parameters that are
        parallelized on, as well as all over all date incremnts as specified in time_increment. For example, if a user
        is calls client.get_asset_metrics(assets="btc,eth,algo", ...).parallel.get_parallel_datacollections() this will
        return three data collections split by asset. If they instead called
        client.get_asset_metrics(assets="btc,eth,algo", metrics="volume_reported_spot_usd_1d", "volume_trusted_spot_usd_1d").parallel(paralellize_on=["metrics", "assets']).get_parallel_datacollections()
        this would instead create 6 data collections combinations. There is also a possible time increment, so if the
        user did client.get_asset_metrics(assets="btc,eth,algo", metrics="volume_reported_spot_usd_1d", "volume_trusted_spot_usd_1d", start_time="2023-01-01", end_time="2023-02-01).parallel(paralellize_on=["metrics", "assets'], time_increment=timedelta(weeks=2)).get_parallel_datacollections()
        it would create 12 data collections total, since it would split it by the 2 week increment as well.
        :return: List[DataCollection] all combinations of DataCollections based on the parallelized parameters and
        time increment.
        """
        data_collections = []
        if len(self._parallelize_on) == 1:
            query_items = self._url_params[self._parallelize_on[0]]
            if isinstance(query_items, str):
                query_items = query_items.split(",")
            for item in query_items:  # type: ignore
                new_params = self._url_params.copy()
                new_params[self._parallelize_on[0]] = item
                new_data_collection = DataCollection(data_retrieval_function=self._data_retrieval_function,
                                                     endpoint=self._endpoint,
                                                     url_params=new_params,
                                                     csv_export_supported=True)
                data_collections.append(new_data_collection)
            data_collections = self._add_time_dimension_to_data_collections(data_collections=data_collections)
            return data_collections

        query_items_dict = {}
        for param in self._parallelize_on:
            if isinstance(self._url_params.get(param), str) and "," in self._url_params.get(param):  # type: ignore
                query_items_dict[param] = self._url_params[param].split(",")  # type: ignore
            else:
                query_items_dict[param] = self._url_params[param]

        combinations = []
        keys = list(query_items_dict.keys())
        for values_combo in itertools.product(*query_items_dict.values()):
            combinations.append(dict(zip(keys, values_combo)))

        for combo in combinations:
            new_params = self._url_params.copy()
            new_params.update(combo)
            new_data_collection = DataCollection(data_retrieval_function=self._data_retrieval_function,
                                                 endpoint=self._endpoint,
                                                 url_params=new_params,
                                                 csv_export_supported=True)
            data_collections.append(new_data_collection)

        data_collections = self._add_time_dimension_to_data_collections(data_collections=data_collections)
        return data_collections

    def _add_time_dimension_to_data_collections(self, data_collections: List[DataCollection]) -> List[DataCollection]:
        """
        Helper function to help create all possible combinations of time + parallelized parameters. Takes a list of
        of data collections and returns a larger a list of dataframe over the time range.
        :param data_collections: List[DataCollection] list of data collections to be expanded
        :return: List[DataCollections] All combinations of the original data collections, over the specified time_increment
        """
        def generate_date_ranges(start: datetime, end: datetime, increment: Union[timedelta, relativedelta]) -> Generator[Tuple[datetime, datetime], None, None]:
            current = start
            if isinstance(increment, timedelta):
                while current < end:
                    next_date = current + increment
                    if next_date > end:
                        next_date = end
                    yield (current, next_date)
                    current = next_date
            elif isinstance(increment, relativedelta):
                while current < end:
                    next_date = current + increment
                    if next_date > end:
                        next_date = end
                    yield (current, next_date)
                    current = next_date
            else:
                raise ValueError("Unsupported increment type")

        if not self._time_increment:
            return data_collections
        if not self._url_params.get("start_time"):
            raise ValueError("No start_time specified, cannot use time_increment feature")

        start_time = self.parse_date(cast(datetime, self._url_params.get("start_time")))
        end_time: datetime = self.parse_date(cast(datetime, self._url_params.get("end_time"))) if self._url_params.get(
            "end_time") else datetime.today()
        full_data_collections = []
        for start, end in generate_date_ranges(start_time, end_time, increment=self._time_increment):
            for data_collection in data_collections:
                new_data_collection = deepcopy(data_collection)
                new_data_collection._url_params.update({"start_time": start, "end_time": end})
                full_data_collections.append(new_data_collection)
        return full_data_collections

    def to_list(self) -> List[Dict[str, Any]]:
        data_collections = self.get_parallel_datacollections()
        total_tasks = len(data_collections)
        with self._executor(max_workers=self._max_workers) as processor:
            if self._progress_bar:
                combined_data = list(
                    tqdm(processor.map(ParallelDataCollection._helper_to_list, data_collections), total=total_tasks,
                         desc="Converting to List"))
            else:
                combined_data = processor.map(ParallelDataCollection._helper_to_list, data_collections)  # type: ignore
        combined_list = list(itertools.chain.from_iterable(combined_data))
        return combined_list

    def to_dataframe(
        self,
        header: Optional[List[str]] = None,
        dtype_mapper: Optional[Dict[str, Any]] = None,
        optimize_pandas_types: Optional[bool] = True,
    ) -> DataFrameType:

        def group_and_merge(dfs: List[pd.DataFrame]) -> pd.DataFrame:
            """
            This function is a helper to merge the kind of dataframes we are dealing with. All the dataframes with
            common columns are merged first, then they are all merged
            :param dfs: List of dataframes
            :return: DataFrame that combines the full list into one
            """
            grouped_dfs = defaultdict(list)

            for df in dfs:
                key = tuple(df.columns)
                grouped_dfs[key].append(df)

            concatenated_dfs = [pd.concat(group, axis=0) for group in grouped_dfs.values()]

            result = concatenated_dfs[0]
            for df in concatenated_dfs[1:]:
                result = pd.merge(result, df, on=['time', self._get_first_param_from_endpoint().rstrip("s")], how='outer')
            return result

        data_collections = self.get_parallel_datacollections()
        with self._executor(max_workers=self._max_workers) as processor:
            if self._progress_bar:
                combined_dataframes = list(tqdm(processor.map(ParallelDataCollection._helper_to_dataframe, data_collections), total=len(data_collections), desc="Exporting to dataframe type"))
            else:
                combined_dataframes = list(processor.map(ParallelDataCollection._helper_to_dataframe, data_collections))

        if len(self._parallelize_on) > 1 or (len(self._parallelize_on) == 1 and self._get_first_param_from_endpoint() != self._parallelize_on[0]):
            combined_df = group_and_merge(combined_dataframes)
        else:
            combined_df = pd.concat(combined_dataframes, axis=0)
        combined_df.reset_index(drop=True, inplace=True)
        return combined_df

    def export_to_csv_files(
        self,
        data_directory: Optional[str] = None,
        columns_to_store: Optional[List[str]] = None,
        compress: bool = False,
    ) -> None:
        """
        This function will export the data requested to several csvs, based on the `parallize_on` attribute of the
        parent class, for example:
        client.get_market_trades("coinbase-eth-btc-spot,coinbase-eth-usdc-spot,coinbase-1inch-btc-spot") will create a
        file each like coinbase-eth-btc.csv, coinbase-eth-usdc-spot.csv, coinbase-1inch-btc-spot.csv
        :param data_directory: str path to directory where files should be dropped
        :param columns_to_store: List[str] columns to store
        :param compress: bool whether or not to compress to tar files
        """
        if data_directory is None:
            data_directory = "."
        data_collections = self.get_parallel_datacollections()
        data_directorys = [data_directory] * len(data_collections)
        columns_to_store_args = [columns_to_store] * len(data_collections)
        compress_args = [compress] * len(data_collections)
        total_tasks = len(data_collections)
        with self._executor(max_workers=self._max_workers) as processor:
            if self._progress_bar:
                list(tqdm(processor.map(self._helper_to_csv, data_collections, data_directorys, columns_to_store_args, compress_args), total=total_tasks, desc="Exporting to CSV"))
            else:
                processor.map(self._helper_to_csv, data_collections, data_directorys, columns_to_store_args,
                              compress_args)

    def export_to_csv(
        self,
        path_or_bufstr: FilePathOrBuffer = None,
        columns_to_store: Optional[List[str]] = None,
        compress: bool = False,
    ) -> None:
        self.to_dataframe().to_csv(path_or_bufstr)

    def export_to_json(
        self,
        path_or_bufstr: FilePathOrBuffer = None,
        compress: bool = False,
    ) -> Optional[str]:
        data_collections = self.get_parallel_datacollections()
        compress_args = [compress] * len(data_collections)
        with self._executor(max_workers=self._max_workers) as processor:
            if self._progress_bar:
                json_data = list(tqdm(processor.map(ParallelDataCollection._helper_to_json, data_collections, compress_args), total=len(data_collections), desc="Exporting to Json"))
            else:
                json_data = list(processor.map(ParallelDataCollection._helper_to_json, data_collections, compress_args))
        if path_or_bufstr:
            with open(path_or_bufstr, 'w') as file:  # type: ignore
                file.writelines(json_data)
                return None
        else:
            return "".join(json_data)

    def export_to_json_files(
            self,
            data_directory: Optional[str] = None,
            compress: bool = False,
    ) -> None:
        """
        This function will export the data requested to several json, based on the `parallize_on` attribute of the
        parent class, for example:
        client.get_market_trades("coinbase-eth-btc-spot,coinbase-eth-usdc-spot,coinbase-1inch-btc-spot") will create a
        file each like coinbase-eth-btc.json, coinbase-eth-usdc-spot.json, coinbase-1inch-btc-spot.json
        :param data_directory: str path to directory where files should be dropped
        :param columns_to_store: List[str] columns to store
        :param compress: bool whether or not to compress to tar files
        """
        if data_directory is None:
            data_directory = "."
        data_collections = self.get_parallel_datacollections()
        data_directorys = [data_directory] * len(data_collections)
        compress_args = [compress] * len(data_collections)
        with self._executor(max_workers=self._max_workers) as processor:
            if self._progress_bar:
                list(tqdm(processor.map(self._helper_to_json_file, data_collections, data_directorys, compress_args), total=len(data_collections), desc="Exporting to Json Files"))
            else:
                processor.map(self._helper_to_json_file, data_collections, data_directorys,
                              compress_args)

    def _get_parallelize_on(self, parallelize_on: Optional[Union[List[str], str]]) -> List[str]:
        if parallelize_on is None:
            return [self._get_first_param_from_endpoint()]
        if isinstance(parallelize_on, str):
            self._validate_parallelization_param(parallelize_on)
            return [parallelize_on]
        elif isinstance(parallelize_on, list) and all([isinstance(item, str) for item in parallelize_on]):
            [self._validate_parallelization_param(param) for param in parallelize_on]  # type: ignore
            return parallelize_on
        else:
            raise ValueError(f"Parallelize on must be either either a string or a list of strings, instead: {parallelize_on}")

    def _validate_parallelization_param(self, param: str) -> None:
        if param not in self._VALID_PARALLELIZATION_PARAMS:
            raise ValueError(
                f"Invalid parallelization param: {param}, only these values are supported: {self._VALID_PARALLELIZATION_PARAMS}")
        if param not in self._url_params:
            raise ValueError(
                f"Invalid parallelization param: {param}, parallelization param must be in {self._url_params.keys()}")
        if self._url_params.get(param) is None:
            raise ValueError(
                f"Invalid parallelization param: {param}, parallelization param is None, must be a non default value")
        if not isinstance(self._url_params.get(param), list) and len(self._url_params.get(param).split(",")) < 2:  # type: ignore
            raise ValueError(f"Invalid parallelization param: {param} - values must be a list, instead: {self._url_params.get(param)}")

    def _helper_to_csv(self, data_collection: DataCollection,
                       data_directory: str,
                       *args: Any) -> None:
        data_names = []
        for param in self._parallelize_on:
            values = data_collection._url_params.get(param)
            if values:
                if isinstance(values, str) and len(values.split(",")) > 1:
                    data_names.extend(values.split(","))
                else:
                    data_names.append(values)  # type: ignore

        data_name = "_".join(data_names)
        friendly_endpoint_name = data_collection._endpoint.split("/")[-1]
        if self._time_increment and data_collection._url_params.get("start_time"):
            start_time = cast(datetime, data_collection._url_params.get("start_time")).strftime("%Y%m%d-%H%M%S")
            file_name = f"{data_name}_{friendly_endpoint_name}_{start_time}.csv"
        else:
            file_name = f"{data_name}_{friendly_endpoint_name}.csv"
        full_file_path = os.path.join(data_directory, file_name)
        data_collection.export_to_csv(full_file_path, *args)

    def _helper_to_json_file(self, data_collection: DataCollection, data_directory: str, compress: bool = False) -> Optional[str]:
        # endpoint = data_collection._endpoint
        first_param_arg = self._get_first_param_from_endpoint()
        data_name = data_collection._url_params[first_param_arg]
        friendly_endpoint_name = data_collection._endpoint.split("/")[-1]
        file_name = f"{data_name}_{friendly_endpoint_name}.json"
        full_file_path = os.path.join(data_directory, file_name)
        return data_collection.export_to_json(full_file_path, compress)

    def _get_first_param_from_endpoint(self) -> str:
        try:
            if self._endpoint.startswith("blockchain"):
                blockchain_endpoints_split = list(map(lambda e: e.split("/"), [endpoint for endpoint in self._ENDPOINT_FIRST_PARAM_DICT.keys() if endpoint.startswith("blockchain")]))
                self_endpoint_split = self._endpoint.split("/")
                for blockchain_split in blockchain_endpoints_split:
                    if all([keyword in self_endpoint_split or keyword.startswith("{") for keyword in blockchain_split]):
                        endpoint = "/".join(blockchain_split)
                        return self._ENDPOINT_FIRST_PARAM_DICT[endpoint]
                raise ValueError(f"Endpoint: {self._endpoint} not supported for parallel requests")
            else:
                return self._ENDPOINT_FIRST_PARAM_DICT[self._endpoint]
        except KeyError:
            raise ValueError(f"Endpoint: {self._endpoint} not supported for parallel requests")

    @staticmethod
    def _helper_to_json(data_collection: DataCollection, compress: bool = False) -> Optional[str]:
        data = data_collection.export_to_json(compress=compress)
        return data

    @staticmethod
    def _helper_to_dataframe(data_collection: DataCollection) -> pd.DataFrame:
        return data_collection.to_dataframe()

    @staticmethod
    def _helper_to_list(data_collection: DataCollection) -> List[Dict[str, Any]]:
        return data_collection.to_list()

    @staticmethod
    def parse_date(date_input: Union[datetime, date, str]) -> datetime:
        if isinstance(date_input, datetime):
            return date_input
        if isinstance(date_input, date):
            return datetime(date_input.year, date_input.month, date_input.day)
        formats = [
            "%Y-%m-%dT%H:%M:%SZ",
            "%Y-%m-%dT%H:%M:%S.%fZ",
            "%Y-%m-%dT%H:%M:%S.%fZ",
            "%Y-%m-%d",
            "%Y%m%d"
        ]
        for fmt in formats:
            try:
                return datetime.strptime(date_input, fmt)
            except ValueError:
                continue
        raise ValueError(f"Unrecognized date format for string: {date_input}")



import datetime

import pytest

from coinmetrics.api_client import CoinMetricsClient
import os

client = CoinMetricsClient(str(os.environ.get("CM_API_KEY")))
cm_api_key_set = os.environ.get("CM_API_KEY") is not None
REASON_TO_SKIP = "Need to set CM_API_KEY as an env var in order to run this test"

print("CM_API_KEY is set - tests will run") if cm_api_key_set else print(
    "CM_API_KEY not set, tests will not run"
)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_as_list_works_like_first_page_short_query() -> None:
    """
    In theory, on a small query that works on a small number of results should have equivalent results for .first_page()
    and also _to_list()
    """
    test_time = datetime.datetime(year=2022, month=2, day=1)
    assets = ["eth", "algo", "btc"]
    data_first_page = client.get_asset_metrics(
        assets=assets,
        metrics="SplyFF",
        limit_per_asset=1,
        start_time=test_time,
        end_time=test_time,
    ).first_page()
    data_list = client.get_asset_metrics(
        assets=assets,
        metrics="SplyFF",
        limit_per_asset=1,
        start_time=test_time,
        end_time=test_time,
    ).to_list()
    assert data_list == data_first_page


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_as_list_doesnt_work_like_first_page_long_query() -> None:
    """
    In theory, on a long query, that will have multiple pages of response data should have very different results
    as first page
    """
    start_time = datetime.datetime(year=2021, month=1, day=1)
    end_time = datetime.datetime(year=2022, month=2, day=1)
    assets = ["eth", "algo", "btc"]
    data_first_page = client.get_asset_metrics(
        assets=assets,
        metrics="SplyFF",
        limit_per_asset=100,
        start_time=start_time,
        end_time=end_time,
    ).first_page()
    data_list = client.get_asset_metrics(
        assets=assets,
        metrics="SplyFF",
        limit_per_asset=100,
        start_time=start_time,
        end_time=end_time,
    ).to_list()
    assert data_list != data_first_page


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_as_list_gets_all_data() -> None:
    """
    This tests that .to_list() gets the same amount and the same data as _to_dataframe()
    """
    start_time = datetime.datetime(year=2021, month=1, day=1)
    end_time = datetime.datetime(year=2022, month=2, day=1)
    assets = ["eth"]
    data_df = client.get_asset_metrics(
        assets=assets,
        metrics="ReferenceRateUSD",
        start_time=start_time,
        end_time=end_time,
    ).to_dataframe()
    data_list = client.get_asset_metrics(
        assets=assets,
        metrics="ReferenceRateUSD",
        start_time=start_time,
        end_time=end_time,
    ).to_list()
    assert len(data_df) == len(data_list)


if __name__ == "__main__":
    pytest.main()



# type: ignore
import datetime
from datetime import timedelta
import os

import dateutil.relativedelta
import pandas as pd
from coinmetrics.api_client import CoinMetricsClient
from coinmetrics._data_collection import DataCollection, ParallelDataCollection
from copy import deepcopy
from enum import Enum
from typing import Tuple



client = CoinMetricsClient(os.environ.get("CM_API_KEY"))

def compare_performance_data_collections(data_collection: DataCollection, max_workers: int = 10) -> None:
    """
    This function will compare
    """
    compare_export_to_csv(data_collection)
    compare_export_to_df(data_collection)
    compare_export_to_csv_normal_parallel_multiple_csvs(data_collection)
    compare_export_to_json_normal_parallel_multiple_jsons(data_collection)


def compare_export_to_csv(data_collection: DataCollection, max_workers: int = 10) -> None:
    start_time_normal = datetime.datetime.now()
    _normal_trades = deepcopy(data_collection).export_to_csv("test.csv")
    print(f"normal export to csv took: {datetime.datetime.now() - start_time_normal}")
    start_time_parallel = datetime.datetime.now()
    _parallel_trades = deepcopy(data_collection).parallel(max_workers=max_workers).export_to_csv("test_parallel.csv")
    print(f"parallel export to csv took: {datetime.datetime.now() - start_time_parallel}")


def compare_export_to_df(data_collection: DataCollection, max_workers: int = 10) -> Tuple[pd.DataFrame, pd.DataFrame]:
    start_time_normal = datetime.datetime.now()
    normal_df = deepcopy(data_collection).to_dataframe()
    print(f"normal export to dataframe took: {datetime.datetime.now() - start_time_normal}")
    start_time_parallel = datetime.datetime.now()
    parallel_df = deepcopy(data_collection).parallel(max_workers=max_workers).to_dataframe()
    print(f"parallel export to dataframe took: {datetime.datetime.now() - start_time_parallel}")
    return normal_df, parallel_df


def compare_export_to_list(data_collection: DataCollection, max_workers: int = 10):
    start_time_normal = datetime.datetime.now()
    normal_df = deepcopy(data_collection).to_list()
    print(f"normal export to list took: {datetime.datetime.now() - start_time_normal}")
    start_time_parallel = datetime.datetime.now()
    parallel_df = deepcopy(data_collection).parallel(max_workers=max_workers).to_list()
    print(f"parallel export to list took: {datetime.datetime.now() - start_time_parallel}")

def compare_export_to_csv_normal_parallel_multiple_csvs(data_collection: DataCollection):
    start_time_normal = datetime.datetime.now()
    normal_df = deepcopy(data_collection).export_to_csv("testsingle.csv")
    print(f"normal export to single csv file took: {datetime.datetime.now() - start_time_normal}")
    start_time_parallel = datetime.datetime.now()
    parallel_df = deepcopy(data_collection).parallel().export_to_csv_files()
    print(f"parallel export to many csv files took: {datetime.datetime.now() - start_time_parallel}")
    return normal_df, parallel_df

def compare_export_to_json_normal_parallel_multiple_jsons(data_collection: DataCollection, max_workers: int = 10):
    start_time_normal = datetime.datetime.now()
    normal_df = deepcopy(data_collection).export_to_csv("testsingle.json")
    print(f"normal export to single json file took: {datetime.datetime.now() - start_time_normal}")
    start_time_parallel = datetime.datetime.now()
    parallel_df = deepcopy(data_collection).parallel(max_workers=max_workers).export_to_json_files()
    print(f"parallel export to many json files took: {datetime.datetime.now() - start_time_parallel}")
    return normal_df, parallel_df

def parallel_time_analysis(data_collection: DataCollection) -> None:
    if not os.path.exists("data"):
        os.mkdir("data")
    start_time = datetime.datetime.now()
    deepcopy(data_collection).parallel().export_to_csv("examplecsv.csv")
    print(f"Parallel export to csv took: {datetime.datetime.now() - start_time}")
    start_time = datetime.datetime.now()
    deepcopy(data_collection).parallel().export_to_csv_files("./data")
    print(f"Parallel export to many csv took: {datetime.datetime.now() - start_time}")
    start_time = datetime.datetime.now()
    deepcopy(data_collection).parallel().export_to_json()
    print(f"Parallel export to json: took {datetime.datetime.now() - start_time}")
    start_time = datetime.datetime.now()
    deepcopy(data_collection).parallel().export_to_json_files("./data")
    print(f"Parallel export to many json files took: {datetime.datetime.now() - start_time}")
    start_time = datetime.datetime.now()
    deepcopy(data_collection).parallel().to_list()
    print(f"Parallel export to list took: {datetime.datetime.now() - start_time}")
    start_time = datetime.datetime.now()
    deepcopy(data_collection).parallel().to_dataframe()
    print(f"Parallel export to dataframe took: {datetime.datetime.now() - start_time}")


def normal_time_analysis(data_collection: DataCollection) -> None:
    if not os.path.exists("data"):
        os.mkdir("data")
    start_time = datetime.datetime.now()
    deepcopy(data_collection).export_to_csv("examplecsv.csv")
    print(f"Normal export to csv took: {datetime.datetime.now() - start_time}")
    start_time = datetime.datetime.now()
    deepcopy(data_collection).export_to_json()
    print(f"Normal export to json: took {datetime.datetime.now() - start_time}")
    start_time = datetime.datetime.now()
    deepcopy(data_collection).to_list()
    print(f"Normal export to list took: {datetime.datetime.now() - start_time}")
    start_time = datetime.datetime.now()
    deepcopy(data_collection).to_dataframe()
    print(f"Normal export to dataframe took: {datetime.datetime.now() - start_time}")


if __name__ == '__main__':
    client = CoinMetricsClient(os.environ.get("CM_API_KEY"))
    start_time = datetime.datetime.now()
    assets = ["btc", "eth", "algo"]
    client.get_asset_metrics(
        assets=assets,
        metrics="ReferenceRateUSD",
        frequency="1m",
        start_time="2022-03-01",
        end_time="2023-03-01",
        page_size=1000,
        end_inclusive=False).parallel(
        time_increment=dateutil.relativedelta.relativedelta(months=1)).export_to_csv("btcRRs.csv")
    print(f"Time taken parallel: {datetime.datetime.now() - start_time}")
    start_time = datetime.datetime.now()
    client.get_asset_metrics(
        assets=assets,
        metrics="ReferenceRateUSD",
        frequency="1m",
        start_time="2022-03-01",
        end_time="2023-03-01",
        page_size=1000,
        end_inclusive=False).export_to_csv("btcRRsNormal.csv")
    print(f"Time taken normal: {datetime.datetime.now() - start_time}")


# type: ignore
import datetime
from datetime import timedelta
import dateutil.relativedelta
import pandas as pd
import pytest

from coinmetrics.api_client import CoinMetricsClient
import os

client = CoinMetricsClient(str(os.environ.get("CM_API_KEY")))
cm_api_key_set = os.environ.get("CM_API_KEY") is not None
REASON_TO_SKIP = "Need to set CM_API_KEY as an env var in order to run this test"

print("CM_API_KEY is set - tests will run") if cm_api_key_set else print(
    "CM_API_KEY not set, tests will not run"
)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_parallel_market_trades() -> None:
    """
    Basic test to make sure that parallel market trades works over a short horizon
    """
    markets = ['coinbase-eth-usdc-spot', 'coinbase-eth-btc-spot']
    start = datetime.datetime(year=2022, day=1, month=6, hour=0, minute=0, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    end = datetime.datetime(year=2022, day=1, month=6, hour=0, minute=1, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    parallel_trades = client.get_market_trades(markets=markets, start_time=start, end_time=end).parallel().to_list()
    normal_trades = client.get_market_trades(markets=markets, start_time=start, end_time=end).to_list()
    unique_markets = set([trade['market'] for trade in parallel_trades])
    assert len(parallel_trades) > 1
    assert len(parallel_trades) == len(normal_trades)
    for market in markets:
        assert market in unique_markets


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_parallel_market_trades_to_df() -> None:
    """
    Basic test to make sure to_dataframe renders right
    """
    markets = ['coinbase-eth-usdc-spot', 'coinbase-eth-btc-spot']
    start = datetime.datetime(year=2022, day=1, month=6, hour=0, minute=0, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    end = datetime.datetime(year=2022, day=1, month=6, hour=0, minute=1, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    parallel_trades = client.get_market_trades(markets=markets, start_time=start, end_time=end).parallel().to_dataframe()
    normal_trades = client.get_market_trades(markets=markets, start_time=start, end_time=end).to_dataframe()
    assert len(parallel_trades) > 1
    assert len(normal_trades) == len(parallel_trades)
    assert all([market in set(normal_trades['market']) for market in markets])


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_asset_metrics_to_single_csv() -> None:
    metrics = ["volume_reported_spot_usd_1d", "volume_trusted_spot_usd_1d"]
    start = datetime.datetime(year=2022, day=1, month=6, hour=0, minute=0, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    end = datetime.datetime(year=2022, day=4, month=6, hour=3, minute=20, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    normal_data = client.get_asset_metrics(assets="btc", metrics=metrics, start_time=start, end_time=end).to_list()
    _export = client.get_asset_metrics(assets="btc", metrics=metrics, start_time=start, end_time=end).parallel(parallelize_on="metrics").export_to_csv("metrics.csv")
    parallel_data = pd.read_csv("metrics.csv")
    assert len(normal_data) == len(parallel_data)
    assert len(parallel_data) == 4


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_parallel_export_to_csvs() -> None:
    """
    Tests that creates the expected
    """
    expected_files = ["btc_asset-metrics.csv", "eth_asset-metrics.csv", "algo_asset-metrics.csv"]
    start = datetime.datetime(year=2022, day=1, month=6, hour=0, minute=0, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    end = datetime.datetime(year=2022, day=1, month=6, hour=0, minute=20, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    try:
        _export_asset_metrics_csvs = client.get_asset_metrics(assets=["btc", "eth", "algo"],
                                                             metrics="ReferenceRateUSD",
                                                             frequency="1m",
                                                             start_time=start,
                                                             end_time=end).parallel().export_to_csv_files()
        assert all([os.path.exists(file) for file in expected_files])
        assert ([len(pd.read_csv(file)) > 5 for file in expected_files])

    finally:
        for file in expected_files:
            if os.path.exists(file):
                os.remove(file)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_parallel_export_to_single_csv() -> None:
    """
    Tests that creates the expected
    """
    start = datetime.datetime(year=2022, day=1, month=6, hour=0, minute=0, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    end = datetime.datetime(year=2022, day=1, month=6, hour=0, minute=20, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    expected_file = "asset_metrics.csv"
    try:
        export_to_csv = client.get_asset_metrics(assets=["btc", "eth", "algo"],
                                                             metrics="ReferenceRateUSD",
                                                             frequency="1m",
                                                             start_time=start,
                                                             end_time=end).parallel().export_to_csv("asset_metrics.csv")
        normal_data = client.get_asset_metrics(assets=["btc", "eth", "algo"],
                                                             metrics="ReferenceRateUSD",
                                                             frequency="1m",
                                                             start_time=start,
                                                             end_time=end).to_list()
        df_from_csv = pd.read_csv(expected_file)
        assert os.path.exists(expected_file)
        assert len(df_from_csv) > 5
        assert len(normal_data) == len(df_from_csv)
    finally:
        os.remove(expected_file)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_parallel_export_to_json_files() -> None:
    expected_files = ["btc_asset-metrics.json", "eth_asset-metrics.json", "algo_asset-metrics.json"]
    start = datetime.datetime(year=2022, day=1, month=6, hour=0, minute=0, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    end = datetime.datetime(year=2022, day=1, month=6, hour=0, minute=20, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    try:
        _export_asset_metrics_json = client.get_asset_metrics(assets=["btc", "eth", "algo"],
                                                              metrics="ReferenceRateUSD",
                                                              frequency="1m",
                                                              start_time=start,
                                                              end_time=end).parallel().export_to_json_files()
        assert all([os.path.exists(file) for file in expected_files])
        assert ([len(pd.read_json(file, lines=True)) > 5 for file in expected_files])

    finally:
        for file in expected_files:
            if os.path.exists(file):
                os.remove(file)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_parallel_export_to_single_json() -> None:
    start = datetime.datetime(year=2022, day=1, month=6, hour=0, minute=0, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    end = datetime.datetime(year=2022, day=1, month=6, hour=0, minute=20, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    expected_file = "asset_metrics.json"
    try:
        _export_to_json = client.get_asset_metrics(assets=["btc", "eth", "algo"],
                                                 metrics="ReferenceRateUSD",
                                                 frequency="1m",
                                                 start_time=start,
                                                 end_time=end).parallel().export_to_json(expected_file)
        normal_data = client.get_asset_metrics(assets=["btc", "eth", "algo"],
                                               metrics="ReferenceRateUSD",
                                               frequency="1m",
                                               start_time=start,
                                               end_time=end).to_list()
        _df_from_csv = pd.read_json(expected_file, lines=True)
        assert os.path.exists(expected_file)
        assert len(pd.read_json(expected_file, lines=True)) > 5
        assert len(normal_data) == len(pd.read_json(expected_file, lines=True))
    finally:
        os.remove(expected_file)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_parallelize_on_metrics() -> None:
    metrics = ["volume_reported_spot_usd_1d", "volume_trusted_spot_usd_1d"]
    start = datetime.datetime(year=2022, day=1, month=6, hour=0, minute=0, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    end = datetime.datetime(year=2022, day=4, month=6, hour=0, minute=20, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    normal_data = client.get_asset_metrics(assets="btc", metrics=metrics, start_time=start, end_time=end).to_dataframe()
    parallel_data = client.get_asset_metrics(assets="btc", metrics=metrics, start_time=start, end_time=end).parallel(parallelize_on="metrics").to_dataframe()
    assert len(normal_data) == len(parallel_data)
    assert len(parallel_data) == 4


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_parallelize_on_metrics_export_to_csvs() -> None:
    metrics = ["volume_reported_spot_usd_1d", "volume_trusted_spot_usd_1d"]
    expected_files = ["volume_reported_spot_usd_1d_asset-metrics.csv", "volume_trusted_spot_usd_1d_asset-metrics.csv"]
    start = datetime.datetime(year=2022, day=1, month=6, hour=0, minute=0, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    end = datetime.datetime(year=2022, day=4, month=6, hour=0, minute=20, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    try:
        client.get_asset_metrics(assets="btc", metrics=metrics, start_time=start, end_time=end).parallel(
            parallelize_on="metrics").export_to_csv_files()
        for file in expected_files:
            assert os.path.exists(file)
            assert len(pd.read_csv(file)) == 4
    finally:
        for file in expected_files:
            if os.path.exists(file):
                os.remove(file)

@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_parallelize_on_metrics_and_assets_export_to_csvs() -> None:
    metrics = ["volume_reported_spot_usd_1d", "volume_trusted_spot_usd_1d"]
    expected_files = ["btc_volume_reported_spot_usd_1d_asset-metrics.csv",
                      "eth_volume_trusted_spot_usd_1d_asset-metrics.csv",
                      "eth_volume_reported_spot_usd_1d_asset-metrics.csv",
                      "btc_volume_trusted_spot_usd_1d_asset-metrics.csv"]
    start = datetime.datetime(year=2022, day=1, month=6, hour=0, minute=0, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    end = datetime.datetime(year=2022, day=4, month=6, hour=0, minute=20, second=0).strftime("%Y-%m-%dT%H:%M:%SZ")
    try:
        client.get_asset_metrics(assets="btc,eth", metrics=metrics, start_time=start, end_time=end).parallel(
            parallelize_on=["assets", "metrics"]).export_to_csv_files()
        for file in expected_files:
            assert os.path.exists(file)
            assert len(pd.read_csv(file)) == 4
    finally:
        for file in expected_files:
            if os.path.exists(file):
                os.remove(file)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_parallel_to_dataframe_time_increment_month() -> None:
    markets = ['coinbase-eth-usdc-spot']
    test_df_parallel = client.get_market_candles(markets=markets, page_size=1000, start_time="2022-01-01", end_time="2022-03-01", end_inclusive=False).parallel(time_increment=dateutil.relativedelta.relativedelta(months=1)).to_dataframe()
    test_df_normal = client.get_market_candles(markets=markets, page_size=1000, start_time="2022-01-01", end_time="2022-03-01", end_inclusive=False).to_dataframe()
    assert len(test_df_parallel) > 10
    assert len(test_df_normal) == len(test_df_parallel)
    assert set(test_df_parallel.columns) == set(test_df_normal.columns)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_parallel_to_dataframe_time_increment_week() -> None:
    markets = ['coinbase-eth-usdc-spot']
    test_df_parallel = client.get_market_candles(markets=markets, page_size=1000, start_time="2022-01-01", end_time="2022-03-01", end_inclusive=False).parallel(progress_bar=False, time_increment=dateutil.relativedelta.relativedelta(weeks=1)).to_dataframe()
    test_df_normal = client.get_market_candles(markets=markets, page_size=1000, start_time="2022-01-01", end_time="2022-03-01", end_inclusive=False).to_dataframe()
    assert len(test_df_parallel) > 10
    assert len(test_df_normal) == len(test_df_parallel)
    assert set(test_df_parallel.columns) == set(test_df_normal.columns)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_parallel_to_dataframe_time_increment_week_multiple_markets() -> None:
    markets = ['coinbase-eth-usdc-spot', 'coinbase-eth-btc-spot']
    test_df_parallel = client.get_market_candles(markets=markets, page_size=1000, start_time="2022-01-01", end_time="2022-03-01", end_inclusive=False).parallel(progress_bar=False, time_increment=dateutil.relativedelta.relativedelta(weeks=1)).to_dataframe()
    test_df_normal = client.get_market_candles(markets=markets, page_size=1000, start_time="2022-01-01", end_time="2022-03-01", end_inclusive=False).to_dataframe()
    assert len(test_df_parallel) > 10
    assert len(test_df_normal) == len(test_df_parallel)
    assert set(test_df_parallel.columns) == set(test_df_normal.columns)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_parallel_asset_metrics_using_time_increment_enum() -> None:
    metrics = ["ReferenceRateUSD", "AdrAct30dCnt",  "AdrAct7dCnt", "AdrActCnt"]
    test_df_parallel = client.get_asset_metrics(assets="btc", metrics=metrics, page_size=1000, start_time="2022-01-01", end_time="2022-03-01", end_inclusive=False, frequency="1d").parallel(parallelize_on=["metrics"], time_increment=dateutil.relativedelta.relativedelta(months=1)).to_dataframe()
    test_df_normal = client.get_asset_metrics(assets="btc", metrics=metrics, page_size=1000, start_time="2022-01-01", end_time="2022-03-01", end_inclusive=False, frequency="1d").to_dataframe()
    assert len(test_df_parallel) > 10
    assert len(test_df_normal) == len(test_df_parallel)
    assert set(test_df_parallel.columns) == set(test_df_normal.columns)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_export_to_csvs_time_increment_files() -> None:
    assets = ["btc"]
    client.get_asset_metrics(
        assets=assets,
        metrics="ReferenceRateUSD",
        frequency="1m",
        start_time="2023-03-01",
        end_time="2023-05-01",
        page_size=1000,
        end_inclusive=False).parallel(
        time_increment=timedelta(days=1)).export_to_csv("btcRRs.csv")

@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_blockchain_parallel_export() -> None:
    df_temp = client.get_list_of_blocks_v2(
        asset="btc",
        start_time='2023-09-07',
        end_time='2023-09-22',
        page_size=10000,
        end_inclusive=False
    ).parallel(time_increment=timedelta(days=1)).to_dataframe()
    assert len(df_temp) > 2000


if __name__ == '__main__':
    pytest.main()






import datetime

import pytest

from coinmetrics.api_client import CoinMetricsClient
from coinmetrics._models import TransactionTrackerData
import os

client = CoinMetricsClient(str(os.environ.get("CM_API_KEY")))
cm_api_key_set = os.environ.get("CM_API_KEY") is not None
REASON_TO_SKIP = "Need to set CM_API_KEY as an env var in order to run this test"

print("CM_API_KEY is set - tests will run") if cm_api_key_set else print(
    "CM_API_KEY not set, tests will not run"
)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_asset_chains_optional_returns() -> None:
    """
    In the old implementation, this test would fail since 'reorg' wouldn't be in the the df columns
    """
    r = client.get_asset_chains(
        assets='btc',
        start_time='2022-10-22T05:40:00',  # '2022-10-22T00:00:00',
        end_time='2022-10-22T06:40:00')

    print(r)
    df = r.to_dataframe()
    print(df.columns)
    print(df)
    assert 'reorg' in df.columns


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_normal_api_call_asset_chains() -> None:
    """
    Tests that when calling get asset chains, the to_dataframe() will contain those with and without reorgs
    """
    request = client.get_asset_chains(
        assets='btc',
        start_time='2022-10-22T05:40:00',  # '2022-10-22T00:00:00',
        end_time='2022-10-22T10:40:00')
    df = request.to_dataframe()
    assert df.iloc[4].reorg == False
    assert df.iloc[3].reorg == True


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_get_transaction_tracker_df() -> None:
    request = client.get_transaction_tracker(
        'btc',
        start_time='2023-03-30T00:00:00',
        end_time='2023-03-30T00:05:00',
        page_size=200
    )
    df = request.to_dataframe()
    assert all([col in df.columns for col in TransactionTrackerData.get_dataframe_cols()])



if __name__ == '__main__':
    request = client.get_transaction_tracker(
        "btc",
        start_time=datetime.datetime.now() - datetime.timedelta(seconds=10),
        page_size=100
    )
    df = request.to_dataframe()
    print("test!")